/*
 * Capability allocation and management.
 *
 * Copyright (C) 2019, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 *
 * ----
 *
 * Unlike the counting allocator in L4Re, this allocator merely keeps a record
 * of whether a capability has been used or not.
 *
 * See: pkg/l4re-core/l4re/util/include/counting_cap_alloc
 *
 * The intention in providing this functionality is to avoid awkward
 * dependencies caused by bringing in C++ classes and the accompanying
 * boilerplate, standard C++ library and, inevitably, C library functionality.
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/sys/consts.h>
#include <l4/sys/factory.h>
#include <l4/sys/task.h>

#include <string.h>         /* memset */

#include "cap_alloc.h"
#include "semaphore.h"



/* Number of managed capability slots.
   NOTE: Employing a fixed number, but this should be configurable. */

#define num_capabilities  4096

/* Bits per group of slots and number of groups. */

#define bits_per_group    (sizeof(unsigned int) * 8)
#define num_groups        (num_capabilities / bits_per_group)

/* Bitmap of capability slots. */

static unsigned int capabilities[num_groups];

/* Next anticipated free slot. */

static int next_free = 0;

/* Capability index offset of the first slot. */

static int start_offset = -1;



/* Semaphore used if threading has been initialised by ipc_cap_alloc_init. */

l4_cap_idx_t ipc_cap_alloc_semaphore = L4_INVALID_CAP;



/* Reserve and return a capability index. */

static l4_cap_idx_t _ipc_cap_alloc(void)
{
  int current;
  unsigned int *group, selector;

  /* Set the start offset if not set. */

  if (start_offset == -1)
  {
    start_offset = l4re_env()->first_free_cap;

    /* Also clear the capabilities bitmap, although this may have been done. */

    memset(capabilities, 0, num_groups * sizeof(unsigned int));
  }

  /* Start at the anticipated free slot, obtaining each word-sized group. */

  current = next_free;
  group = &capabilities[current / bits_per_group];

  do
  {
    /* Select the bit corresponding to the current slot. */

    selector = 1 << (current % bits_per_group);

    /* Test the bit. If not set, set the bit, update the next free indicator,
       and return the current slot indicator. */

    if (!(*group & selector))
    {
      *group |= selector;
      next_free = (current + 1) % num_capabilities;

      /* Convert the index into a capability value. */

      return (current + start_offset) << L4_CAP_SHIFT;
    }

    /* Move on to the next slot, loading the group if necessary. */

    current = (current + 1) % num_capabilities;

    if (!(current % bits_per_group))
      group = &capabilities[current / bits_per_group];

  } while (current != next_free);

  return L4_INVALID_CAP;
}

l4_cap_idx_t ipc_cap_alloc()
{
  l4_cap_idx_t cap;
  long err;

  /* Allocate a capability in the critical section, then return. */

  err = ipc_semaphore_down(ipc_cap_alloc_semaphore);
  if (err)
    return L4_INVALID_CAP;

  cap = _ipc_cap_alloc();

  ipc_semaphore_up(ipc_cap_alloc_semaphore);

  return cap;
}

/* Release a capability. */

static int _ipc_cap_free(l4_cap_idx_t cap)
{
  int index;
  unsigned int *group, selector;

  if ((start_offset == -1) || l4_is_invalid_cap(cap))
    return 0;

  /* Convert the capability value into an index in the slots. */

  index = cap >> L4_CAP_SHIFT;

  if (index < start_offset)
    return 0;

  index -= start_offset;

  /* Obtain the group of slots and a selector for the capability. */

  group = &capabilities[index / bits_per_group];
  selector = 1 << (index % bits_per_group);

  /* Clear the selected bit. */

  *group &= (~selector);

  return 1;
}

int ipc_cap_free(l4_cap_idx_t cap)
{
  int result;
  long err;

  err = ipc_semaphore_down(ipc_cap_alloc_semaphore);
  if (err)
    return 0;

  result = _ipc_cap_free(cap);

  ipc_semaphore_up(ipc_cap_alloc_semaphore);

  return result;
}

/* Free the indicated capability and unmap it. */

static int _ipc_cap_free_um(l4_cap_idx_t cap)
{
  if (ipc_cap_free(cap))
  {
    l4_task_unmap(L4RE_THIS_TASK_CAP, l4_obj_fpage(cap, 0, L4_FPAGE_RWX), L4_FP_ALL_SPACES);
    return 1;
  }

  return 0;
}

int ipc_cap_free_um(l4_cap_idx_t cap)
{
  int result;
  long err;

  err = ipc_semaphore_down(ipc_cap_alloc_semaphore);
  if (err)
    return 0;

  result = _ipc_cap_free_um(cap);

  ipc_semaphore_up(ipc_cap_alloc_semaphore);

  return result;
}

// vim: tabstop=2 expandtab shiftwidth=2
